
#include "ad++.h"
#include "admacros.h"

#include "raw_type.h"

#include "interface.h"

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  `scatterSum'                                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<int R, class T>
inline void scatterSum(const Section_<R, T>& b, const Section1<T>& a,
                       const Section_<R, int>& s_1) {
  ScatterSum<T> schedule(&b, &a, sizeof(T),
                         Seq1<const Frame*>(&s_1), Seq1<int*>(s_1.dat()),
                         0, 0) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<int R, class T>
inline void scatterSum(const Section_<R, T>& b, const Section2<T>& a,
                       const Section_<R, int>& s_1,
                       const Section_<R, int>& s_2) {
  ScatterSum<T> schedule(&b, &a, sizeof(T),
                         Seq2<const Frame*>(&s_1, &s_2),
                         Seq2<int*>(s_1.dat(), s_2.dat()),
                         0, 0) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<int R, class T>
inline void scatterSum(const Section_<R, T>& b, const Section3<T>& a,
                       const Section_<R, int>& s_1,
                       const Section_<R, int>& s_2,
                       const Section_<R, int>& s_3) {
  ScatterSum<T> schedule(&b, &a, sizeof(T),
                         Seq3<const Frame*>(&s_1, &s_2, &s_3),
                         Seq3<int*>(s_1.dat(), s_2.dat(), s_3.dat()),
                         0, 0) ;

  schedule.execute(b.dat(), a.dat()) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  `scatterProduct'                                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<int R, class T>
inline void scatterProduct(const Section_<R, T>& b, const Section1<T>& a,
                       const Section_<R, int>& s_1) {
  ScatterProduct<T> schedule(&b, &a, sizeof(T),
                         Seq1<const Frame*>(&s_1), Seq1<int*>(s_1.dat()),
                         0, 0) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<int R, class T>
inline void scatterProduct(const Section_<R, T>& b, const Section2<T>& a,
                       const Section_<R, int>& s_1,
                       const Section_<R, int>& s_2) {
  ScatterProduct<T> schedule(&b, &a, sizeof(T),
                         Seq2<const Frame*>(&s_1, &s_2),
                         Seq2<int*>(s_1.dat(), s_2.dat()),
                         0, 0) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<int R, class T>
inline void scatterProduct(const Section_<R, T>& b, const Section3<T>& a,
                       const Section_<R, int>& s_1,
                       const Section_<R, int>& s_2,
                       const Section_<R, int>& s_3) {
  ScatterProduct<T> schedule(&b, &a, sizeof(T),
                         Seq3<const Frame*>(&s_1, &s_2, &s_3),
                         Seq3<int*>(s_1.dat(), s_2.dat(), s_3.dat()),
                         0, 0) ;

  schedule.execute(b.dat(), a.dat()) ;
}



int main(int argc, char* argv[]) {
  AdlibInit(argc, argv) ;

  Procs2 p(3, 2) ;

  // Parameters of 1d arrays (and sections)

  BlockRange s(4, p.dim(1)) ;
  BlockRange x(100, CyclicRange(5, p.dim(0))) ;

  Group q = p / s(3) ;

  // Parameters of 2d arrays

  Range y = CyclicRange(25, p.dim(0)).subrng(10, 1, 2) ;
  Range z = BlockRange(20, p.dim(1)).subrng(10, 5) ;

  // Parameters of 3d arrays

  Range u = StepRange(13, p.dim(0)).subrng(5, 3, 2) ;
  Range v(5) ;
  CyclicRange w(5, p.dim(1)) ;

  ON(p) {
    // int Sum
    
    {
      gprintf("int scatterSum.  1d target, 3d source and subscripts.\n") ;

      Array2<int> apar(s, x, p) ;
      Section1<int> a = apar.sect(s(3), x) ;

      Array3<int> b(u, v, w, p) ;
      Array3<int> s1(u, v, w, p) ;

      Index i(x) ;
      ON(q) {
	WHERE(i) {
	  a(i) = i ;
	} EREWH(i) ;
      } NO(q) ;

      Index j(u), k(v), l(w) ;
      WHERE(j) {
	WHERE(k) {
	  WHERE(l) {
	    b(j, k, l)  = 100 * j + 10 * k + l ;
	    s1(j, k, l) = (13 * j + 11 * k + 7 * l + 3) % 100 ;
	  } EREWH(l) ;
	} EREWH(k) ;
      } EREWH(j) ;

      scatterSum(b, a, s1) ;

      aprintf("%10d\n", a) ;
      gprintf("\n") ;
    }

    {
      gprintf("int scatterSum.  2d target, 2d source and subscripts.\n") ;

      Array2<int> a(y, z, p) ;

      Array2<int> b(y, z, p) ;
      Array2<int> s1(y, z, p), s2(y, z, p) ;

      Index i(y), j(z) ;
      WHERE(i) {
	WHERE(j) {
	  a(i, j) = 10 * i + j ;
	} EREWH(j) ;
      } EREWH(i) ;

      WHERE(i) {
        WHERE(j) {
	  b(i, j)  = 10 * i + j ;
	  s1(i, j) = ((111 * i + 71 * j + 3) % 100) / 10 ;
	  s2(i, j) = ((19 * i + 173 * j + 13) % 100) / 10 ;
        } EREWH(j) ;
      } EREWH(i) ;

      scatterSum(b, a, s1, s2) ;

      aprintf("%10d\n", a) ;
      gprintf("\n") ;
    }

    {
      gprintf("int scatterSum.  3d target, 1d source and subscripts.\n") ;

      Array3<int> a(u, v, w, p) ;

      Array2<int> bpar(s, x, p) ;
      Section1<int> b = bpar.sect(s(3), x) ;
      Array1<int> s1(x, q), s2(x, q), s3(x, q) ;

      Index j(u), k(v), l(w) ;
      WHERE(j) {
	WHERE(k) {
	  WHERE(l) {
	    a(j, k, l)  = 100 * j + 10 * k + l ;
	  } EREWH(l) ;
	} EREWH(k) ;
      } EREWH(j) ;

      Index i(x) ;
      ON(q) {
	WHERE(i) {
	  b(i) = i ;
	  s1(i) = ((73 * i + 3) % 100) / 20 ;
	  s2(i) = ((121 * i + 11) % 100) / 20 ;
	  s3(i) = ((97 * i + 17) % 100) / 20 ;
	} EREWH(i) ;
      } NO(q) ;

      scatterSum(b, a, s1, s2, s3) ;

      aprintf("%10d\n", a) ;
      gprintf("\n") ;
    }

    // float Sum
    
    {
      gprintf("float scatterSum.  1d target, 3d source and subscripts.\n") ;

      Array2<float> apar(s, x, p) ;
      Section1<float> a = apar.sect(s(3), x) ;

      Array3<float> b(u, v, w, p) ;
      Array3<int> s1(u, v, w, p) ;

      Index i(x) ;
      ON(q) {
	WHERE(i) {
	  a(i) = i ;
	} EREWH(i) ;
      } NO(q) ;

      Index j(u), k(v), l(w) ;
      WHERE(j) {
	WHERE(k) {
	  WHERE(l) {
	    b(j, k, l)  = 100 * j + 10 * k + l ;
	    s1(j, k, l) = (13 * j + 11 * k + 7 * l + 3) % 100 ;
	  } EREWH(l) ;
	} EREWH(k) ;
      } EREWH(j) ;

      scatterSum(b, a, s1) ;

      aprintf("%f\n", a) ;
      gprintf("\n") ;
    }

    {
      gprintf("float scatterSum.  2d target, 2d source and subscripts.\n") ;

      Array2<float> a(y, z, p) ;

      Array2<float> b(y, z, p) ;
      Array2<int> s1(y, z, p), s2(y, z, p) ;

      Index i(y), j(z) ;
      WHERE(i) {
	WHERE(j) {
	  a(i, j) = 10 * i + j ;
	} EREWH(j) ;
      } EREWH(i) ;

      WHERE(i) {
        WHERE(j) {
	  b(i, j)  = 10 * i + j ;
	  s1(i, j) = ((111 * i + 71 * j + 3) % 100) / 10 ;
	  s2(i, j) = ((19 * i + 173 * j + 13) % 100) / 10 ;
        } EREWH(j) ;
      } EREWH(i) ;

      scatterSum(b, a, s1, s2) ;

      aprintf("%f\n", a) ;
      gprintf("\n") ;
    }

    {
      gprintf("float scatterSum.  3d target, 1d source and subscripts.\n") ;

      Array3<float> a(u, v, w, p) ;

      Array2<float> bpar(s, x, p) ;
      Section1<float> b = bpar.sect(s(3), x) ;
      Array1<int> s1(x, q), s2(x, q), s3(x, q) ;

      Index j(u), k(v), l(w) ;
      WHERE(j) {
	WHERE(k) {
	  WHERE(l) {
	    a(j, k, l)  = 100 * j + 10 * k + l ;
	  } EREWH(l) ;
	} EREWH(k) ;
      } EREWH(j) ;

      Index i(x) ;
      ON(q) {
	WHERE(i) {
	  b(i) = i ;
	  s1(i) = ((73 * i + 3) % 100) / 20 ;
	  s2(i) = ((121 * i + 11) % 100) / 20 ;
	  s3(i) = ((97 * i + 17) % 100) / 20 ;
	} EREWH(i) ;
      } NO(q) ;

      scatterSum(b, a, s1, s2, s3) ;

      aprintf("%f\n", a) ;
      gprintf("\n") ;
    }

    // double Sum
    
    {
      gprintf("double scatterSum.  1d target, 3d source and subscripts.\n") ;

      Array2<double> apar(s, x, p) ;
      Section1<double> a = apar.sect(s(3), x) ;

      Array3<double> b(u, v, w, p) ;
      Array3<int> s1(u, v, w, p) ;

      Index i(x) ;
      ON(q) {
	WHERE(i) {
	  a(i) = i ;
	} EREWH(i) ;
      } NO(q) ;

      Index j(u), k(v), l(w) ;
      WHERE(j) {
	WHERE(k) {
	  WHERE(l) {
	    b(j, k, l)  = 100 * j + 10 * k + l ;
	    s1(j, k, l) = (13 * j + 11 * k + 7 * l + 3) % 100 ;
	  } EREWH(l) ;
	} EREWH(k) ;
      } EREWH(j) ;

      scatterSum(b, a, s1) ;

      aprintf("%f\n", a) ;
      gprintf("\n") ;
    }

    {
      gprintf("double scatterSum.  2d target, 2d source and subscripts.\n") ;

      Array2<double> a(y, z, p) ;

      Array2<double> b(y, z, p) ;
      Array2<int> s1(y, z, p), s2(y, z, p) ;

      Index i(y), j(z) ;
      WHERE(i) {
	WHERE(j) {
	  a(i, j) = 10 * i + j ;
	} EREWH(j) ;
      } EREWH(i) ;

      WHERE(i) {
        WHERE(j) {
	  b(i, j)  = 10 * i + j ;
	  s1(i, j) = ((111 * i + 71 * j + 3) % 100) / 10 ;
	  s2(i, j) = ((19 * i + 173 * j + 13) % 100) / 10 ;
        } EREWH(j) ;
      } EREWH(i) ;

      scatterSum(b, a, s1, s2) ;

      aprintf("%f\n", a) ;
      gprintf("\n") ;
    }

    {
      gprintf("double scatterSum.  3d target, 1d source and subscripts.\n") ;

      Array3<double> a(u, v, w, p) ;

      Array2<double> bpar(s, x, p) ;
      Section1<double> b = bpar.sect(s(3), x) ;
      Array1<int> s1(x, q), s2(x, q), s3(x, q) ;

      Index j(u), k(v), l(w) ;
      WHERE(j) {
	WHERE(k) {
	  WHERE(l) {
	    a(j, k, l)  = 100 * j + 10 * k + l ;
	  } EREWH(l) ;
	} EREWH(k) ;
      } EREWH(j) ;

      Index i(x) ;
      ON(q) {
	WHERE(i) {
	  b(i) = i ;
	  s1(i) = ((73 * i + 3) % 100) / 20 ;
	  s2(i) = ((121 * i + 11) % 100) / 20 ;
	  s3(i) = ((97 * i + 17) % 100) / 20 ;
	} EREWH(i) ;
      } NO(q) ;

      scatterSum(b, a, s1, s2, s3) ;

      aprintf("%f\n", a) ;
      gprintf("\n") ;
    }

    // COMPLEX Sum
    
    {
      gprintf("COMPLEX scatterSum.  1d target, 3d source and subscripts.\n") ;

      Array2<COMPLEX> apar(s, x, p) ;
      Section1<COMPLEX> a = apar.sect(s(3), x) ;

      Array3<COMPLEX> b(u, v, w, p) ;
      Array3<int> s1(u, v, w, p) ;

      Index i(x) ;
      ON(q) {
	WHERE(i) {
	  a(i) = (int) i ;
	} EREWH(i) ;
      } NO(q) ;

      Index j(u), k(v), l(w) ;
      WHERE(j) {
	WHERE(k) {
	  WHERE(l) {
	    b(j, k, l)  = COMPLEX(0, 100 * j + 10 * k + l) ;
	    s1(j, k, l) = (13 * j + 11 * k + 7 * l + 3) % 100 ;
	  } EREWH(l) ;
	} EREWH(k) ;
      } EREWH(j) ;

      scatterSum(b, a, s1) ;

      Array1<float> re(x, q), im(x, q) ;
      ON(q) {
	WHERE(i) {
	  re(i) = a(i).re() ;
	  im(i) = a(i).im() ;
	} EREWH(i) ;
      } NO(q) ;
      aprintf("(%f %f)\n", re, im) ;
      gprintf("\n") ;
    }

    {
      gprintf("COMPLEX scatterSum.  2d target, 2d source and subscripts.\n") ;

      Array2<COMPLEX> a(y, z, p) ;

      Array2<COMPLEX> b(y, z, p) ;
      Array2<int> s1(y, z, p), s2(y, z, p) ;

      Index i(y), j(z) ;
      WHERE(i) {
	WHERE(j) {
	  a(i, j) = 10 * i + j ;
	} EREWH(j) ;
      } EREWH(i) ;

      WHERE(i) {
        WHERE(j) {
	  b(i, j)  = COMPLEX(0, 10 * i + j) ;
	  s1(i, j) = ((111 * i + 71 * j + 3) % 100) / 10 ;
	  s2(i, j) = ((19 * i + 173 * j + 13) % 100) / 10 ;
        } EREWH(j) ;
      } EREWH(i) ;

      scatterSum(b, a, s1, s2) ;

      Array2<float> re(y, z, p), im(y, z, p) ;
      WHERE(i) {
	WHERE(j) {
	  re(i, j) = a(i, j).re() ;
	  im(i, j) = a(i, j).im() ;
	} EREWH(j) ;
      } EREWH(i) ;
      aprintf("(%f %f)\n", re, im) ;
      gprintf("\n") ;
    }

    {
      gprintf("COMPLEX scatterSum.  3d target, 1d source and subscripts.\n") ;

      Array3<COMPLEX> a(u, v, w, p) ;

      Array2<COMPLEX> bpar(s, x, p) ;
      Section1<COMPLEX> b = bpar.sect(s(3), x) ;
      Array1<int> s1(x, q), s2(x, q), s3(x, q) ;

      Index j(u), k(v), l(w) ;
      WHERE(j) {
	WHERE(k) {
	  WHERE(l) {
	    a(j, k, l)  = 100 * j + 10 * k + l ;
	  } EREWH(l) ;
	} EREWH(k) ;
      } EREWH(j) ;

      Index i(x) ;
      ON(q) {
	WHERE(i) {
	  b(i) = COMPLEX(0, i) ;
	  s1(i) = ((73 * i + 3) % 100) / 20 ;
	  s2(i) = ((121 * i + 11) % 100) / 20 ;
	  s3(i) = ((97 * i + 17) % 100) / 20 ;
	} EREWH(i) ;
      } NO(q) ;

      scatterSum(b, a, s1, s2, s3) ;

      Array3<float> re(u, v, w, p), im(u, v, w, p) ;
      WHERE(j) {
	WHERE(k) {
	  WHERE(l) {
	    re(j, k, l) = a(j, k, l).re() ;
	    im(j, k, l) = a(j, k, l).im() ;
	  } EREWH(l) ;
	} EREWH(k) ;
      } EREWH(j) ;
      aprintf("(%f %f)\n", re, im) ;
      gprintf("\n") ;
    }

    // int Product
    
    {
      gprintf("int scatterProduct.  1d target, 3d source and subscripts.\n") ;

      Array2<int> apar(s, x, p) ;
      Section1<int> a = apar.sect(s(3), x) ;

      Array3<int> b(u, v, w, p) ;
      Array3<int> s1(u, v, w, p) ;

      Index i(x) ;
      ON(q) {
	WHERE(i) {
	  a(i) = i ;
	} EREWH(i) ;
      } NO(q) ;

      Index j(u), k(v), l(w) ;
      WHERE(j) {
	WHERE(k) {
	  WHERE(l) {
	    b(j, k, l)  = (100 * j + 10 * k + l + 9) / 10 ;
	    s1(j, k, l) = (13 * j + 11 * k + 7 * l + 3) % 100 ;
	  } EREWH(l) ;
	} EREWH(k) ;
      } EREWH(j) ;

      scatterProduct(b, a, s1) ;

      aprintf("%10d\n", a) ;
      gprintf("\n") ;
    }

    {
      gprintf("int scatterProduct.  2d target, 2d source and subscripts.\n") ;

      Array2<int> a(y, z, p) ;

      Array2<int> b(y, z, p) ;
      Array2<int> s1(y, z, p), s2(y, z, p) ;

      Index i(y), j(z) ;
      WHERE(i) {
	WHERE(j) {
	  a(i, j) = 10 * i + j ;
	} EREWH(j) ;
      } EREWH(i) ;

      WHERE(i) {
        WHERE(j) {
	  b(i, j)  = (10 * i + j + 9) / 10 ;
	  s1(i, j) = ((111 * i + 71 * j + 3) % 100) / 10 ;
	  s2(i, j) = ((19 * i + 173 * j + 13) % 100) / 10 ;
        } EREWH(j) ;
      } EREWH(i) ;

      scatterProduct(b, a, s1, s2) ;

      aprintf("%10d\n", a) ;
      gprintf("\n") ;
    }

    {
      gprintf("int scatterProduct.  3d target, 1d source and subscripts.\n") ;

      Array3<int> a(u, v, w, p) ;

      Array2<int> bpar(s, x, p) ;
      Section1<int> b = bpar.sect(s(3), x) ;
      Array1<int> s1(x, q), s2(x, q), s3(x, q) ;

      Index j(u), k(v), l(w) ;
      WHERE(j) {
	WHERE(k) {
	  WHERE(l) {
	    a(j, k, l)  = 100 * j + 10 * k + l ;
	  } EREWH(l) ;
	} EREWH(k) ;
      } EREWH(j) ;

      Index i(x) ;
      ON(q) {
	WHERE(i) {
	  b(i) = (i + 9) / 10 ;
	  s1(i) = ((73 * i + 3) % 100) / 20 ;
	  s2(i) = ((121 * i + 11) % 100) / 20 ;
	  s3(i) = ((97 * i + 17) % 100) / 20 ;
	} EREWH(i) ;
      } NO(q) ;

      scatterProduct(b, a, s1, s2, s3) ;

      aprintf("%10d\n", a) ;
      gprintf("\n") ;
    }

    // float Product
    
    {
      gprintf("float scatterProduct.  1d target, 3d source and subscripts.\n") ;

      Array2<float> apar(s, x, p) ;
      Section1<float> a = apar.sect(s(3), x) ;

      Array3<float> b(u, v, w, p) ;
      Array3<int> s1(u, v, w, p) ;

      Index i(x) ;
      ON(q) {
	WHERE(i) {
	  a(i) = i ;
	} EREWH(i) ;
      } NO(q) ;

      Index j(u), k(v), l(w) ;
      WHERE(j) {
	WHERE(k) {
	  WHERE(l) {
	    b(j, k, l)  = (100 * j + 10 * k + l + 9) / 10 ;
	    s1(j, k, l) = (13 * j + 11 * k + 7 * l + 3) % 100 ;
	  } EREWH(l) ;
	} EREWH(k) ;
      } EREWH(j) ;

      scatterProduct(b, a, s1) ;

      aprintf("%f\n", a) ;
      gprintf("\n") ;
    }

    {
      gprintf("float scatterProduct.  2d target, 2d source and subscripts.\n") ;

      Array2<float> a(y, z, p) ;

      Array2<float> b(y, z, p) ;
      Array2<int> s1(y, z, p), s2(y, z, p) ;

      Index i(y), j(z) ;
      WHERE(i) {
	WHERE(j) {
	  a(i, j) = 10 * i + j ;
	} EREWH(j) ;
      } EREWH(i) ;

      WHERE(i) {
        WHERE(j) {
	  b(i, j)  = (10 * i + j + 9) / 10 ;
	  s1(i, j) = ((111 * i + 71 * j + 3) % 100) / 10 ;
	  s2(i, j) = ((19 * i + 173 * j + 13) % 100) / 10 ;
        } EREWH(j) ;
      } EREWH(i) ;

      scatterProduct(b, a, s1, s2) ;

      aprintf("%f\n", a) ;
      gprintf("\n") ;
    }

    {
      gprintf("float scatterProduct.  3d target, 1d source and subscripts.\n") ;

      Array3<float> a(u, v, w, p) ;

      Array2<float> bpar(s, x, p) ;
      Section1<float> b = bpar.sect(s(3), x) ;
      Array1<int> s1(x, q), s2(x, q), s3(x, q) ;

      Index j(u), k(v), l(w) ;
      WHERE(j) {
	WHERE(k) {
	  WHERE(l) {
	    a(j, k, l)  = 100 * j + 10 * k + l ;
	  } EREWH(l) ;
	} EREWH(k) ;
      } EREWH(j) ;

      Index i(x) ;
      ON(q) {
	WHERE(i) {
	  b(i) = (i + 9) / 10 ;
	  s1(i) = ((73 * i + 3) % 100) / 20 ;
	  s2(i) = ((121 * i + 11) % 100) / 20 ;
	  s3(i) = ((97 * i + 17) % 100) / 20 ;
	} EREWH(i) ;
      } NO(q) ;

      scatterProduct(b, a, s1, s2, s3) ;

      aprintf("%f\n", a) ;
      gprintf("\n") ;
    }

    // double Product
    
    {
      gprintf("double scatterProduct.  1d target, 3d source and subscripts.\n") ;

      Array2<double> apar(s, x, p) ;
      Section1<double> a = apar.sect(s(3), x) ;

      Array3<double> b(u, v, w, p) ;
      Array3<int> s1(u, v, w, p) ;

      Index i(x) ;
      ON(q) {
	WHERE(i) {
	  a(i) = i ;
	} EREWH(i) ;
      } NO(q) ;

      Index j(u), k(v), l(w) ;
      WHERE(j) {
	WHERE(k) {
	  WHERE(l) {
	    b(j, k, l)  = (100 * j + 10 * k + l + 9) / 10 ;
	    s1(j, k, l) = (13 * j + 11 * k + 7 * l + 3) % 100 ;
	  } EREWH(l) ;
	} EREWH(k) ;
      } EREWH(j) ;

      scatterProduct(b, a, s1) ;

      aprintf("%f\n", a) ;
      gprintf("\n") ;
    }

    {
      gprintf("double scatterProduct.  2d target, 2d source and subscripts.\n") ;

      Array2<double> a(y, z, p) ;

      Array2<double> b(y, z, p) ;
      Array2<int> s1(y, z, p), s2(y, z, p) ;

      Index i(y), j(z) ;
      WHERE(i) {
	WHERE(j) {
	  a(i, j) = 10 * i + j ;
	} EREWH(j) ;
      } EREWH(i) ;

      WHERE(i) {
        WHERE(j) {
	  b(i, j)  = (10 * i + j + 9) / 10 ;
	  s1(i, j) = ((111 * i + 71 * j + 3) % 100) / 10 ;
	  s2(i, j) = ((19 * i + 173 * j + 13) % 100) / 10 ;
        } EREWH(j) ;
      } EREWH(i) ;

      scatterProduct(b, a, s1, s2) ;

      aprintf("%f\n", a) ;
      gprintf("\n") ;
    }

    {
      gprintf("double scatterProduct.  3d target, 1d source and subscripts.\n") ;

      Array3<double> a(u, v, w, p) ;

      Array2<double> bpar(s, x, p) ;
      Section1<double> b = bpar.sect(s(3), x) ;
      Array1<int> s1(x, q), s2(x, q), s3(x, q) ;

      Index j(u), k(v), l(w) ;
      WHERE(j) {
	WHERE(k) {
	  WHERE(l) {
	    a(j, k, l)  = 100 * j + 10 * k + l ;
	  } EREWH(l) ;
	} EREWH(k) ;
      } EREWH(j) ;

      Index i(x) ;
      ON(q) {
	WHERE(i) {
	  b(i) = (i + 9) / 10 ;
	  s1(i) = ((73 * i + 3) % 100) / 20 ;
	  s2(i) = ((121 * i + 11) % 100) / 20 ;
	  s3(i) = ((97 * i + 17) % 100) / 20 ;
	} EREWH(i) ;
      } NO(q) ;

      scatterProduct(b, a, s1, s2, s3) ;

      aprintf("%f\n", a) ;
      gprintf("\n") ;
    }

    // COMPLEX Product
    
    {
      gprintf("COMPLEX scatterProduct.  1d target, 3d source and subscripts.\n") ;

      Array2<COMPLEX> apar(s, x, p) ;
      Section1<COMPLEX> a = apar.sect(s(3), x) ;

      Array3<COMPLEX> b(u, v, w, p) ;
      Array3<int> s1(u, v, w, p) ;

      Index i(x) ;
      ON(q) {
	WHERE(i) {
	  a(i) = (int) i ;
	} EREWH(i) ;
      } NO(q) ;

      Index j(u), k(v), l(w) ;
      WHERE(j) {
	WHERE(k) {
	  WHERE(l) {
	    b(j, k, l)  = COMPLEX(0, (100 * j + 10 * k + l + 9) / 10) ;
	    s1(j, k, l) = (13 * j + 11 * k + 7 * l + 3) % 100 ;
	  } EREWH(l) ;
	} EREWH(k) ;
      } EREWH(j) ;

      scatterProduct(b, a, s1) ;

      Array1<float> re(x, q), im(x, q) ;
      ON(q) {
        WHERE(i) {
          re(i) = a(i).re() ;
          im(i) = a(i).im() ;
        } EREWH(i) ;
      } NO(q) ;
      aprintf("(%f %f)\n", re, im) ;
      gprintf("\n") ;
    }

    {
      gprintf("COMPLEX scatterProduct.  2d target, 2d source and subscripts.\n") ;

      Array2<COMPLEX> a(y, z, p) ;

      Array2<COMPLEX> b(y, z, p) ;
      Array2<int> s1(y, z, p), s2(y, z, p) ;

      Index i(y), j(z) ;
      WHERE(i) {
	WHERE(j) {
	  a(i, j) = 10 * i + j ;
	} EREWH(j) ;
      } EREWH(i) ;

      WHERE(i) {
        WHERE(j) {
	  b(i, j)  = COMPLEX(0, (10 * i + j + 9) / 10) ;
	  s1(i, j) = ((111 * i + 71 * j + 3) % 100) / 10 ;
	  s2(i, j) = ((19 * i + 173 * j + 13) % 100) / 10 ;
        } EREWH(j) ;
      } EREWH(i) ;

      scatterProduct(b, a, s1, s2) ;

      Array2<float> re(y, z, p), im(y, z, p) ;
      WHERE(i) {
	WHERE(j) {
	  re(i, j) = a(i, j).re() ;
	  im(i, j) = a(i, j).im() ;
	} EREWH(j) ;
      } EREWH(i) ;
      aprintf("(%f %f)\n", re, im) ;
      gprintf("\n") ;
    }

    {
      gprintf("COMPLEX scatterProduct.  3d target, 1d source and subscripts.\n") ;

      Array3<COMPLEX> a(u, v, w, p) ;

      Array2<COMPLEX> bpar(s, x, p) ;
      Section1<COMPLEX> b = bpar.sect(s(3), x) ;
      Array1<int> s1(x, q), s2(x, q), s3(x, q) ;

      Index j(u), k(v), l(w) ;
      WHERE(j) {
	WHERE(k) {
	  WHERE(l) {
	    a(j, k, l)  = 100 * j + 10 * k + l ;
	  } EREWH(l) ;
	} EREWH(k) ;
      } EREWH(j) ;

      Index i(x) ;
      ON(q) {
	WHERE(i) {
	  b(i) = COMPLEX(0, (i + 9) / 10) ;
	  s1(i) = ((73 * i + 3) % 100) / 20 ;
	  s2(i) = ((121 * i + 11) % 100) / 20 ;
	  s3(i) = ((97 * i + 17) % 100) / 20 ;
	} EREWH(i) ;
      } NO(q) ;

      scatterProduct(b, a, s1, s2, s3) ;

      Array3<float> re(u, v, w, p), im(u, v, w, p) ;
      WHERE(j) {
	WHERE(k) {
	  WHERE(l) {
	    re(j, k, l) = a(j, k, l).re() ;
	    im(j, k, l) = a(j, k, l).im() ;
	  } EREWH(l) ;
	} EREWH(k) ;
      } EREWH(j) ;
      aprintf("(%f %f)\n", re, im) ;
      gprintf("\n") ;
    }

  } NO(p) ;

  AdlibFinalize() ;
}


